@@ -16,13 +16,14 @@ const Signals = imports.signals;
 const St = imports.gi.St;
 const Main = imports.ui.main;
 const Params = imports.misc.params;
-const Tweener = imports.ui.tweener;
 const EdgeFlip = imports.ui.edgeFlip;
 const HotCorner = imports.ui.hotCorner;
 const DeskletManager = imports.ui.deskletManager;
 const Panel = imports.ui.panel;
 const StartupAnimation = imports.ui.startupAnimation;
 
+var KEYBOARD_FADE_TIME = 150;
+
 function isPopupMetaWindow(actor) {
     switch(actor.meta_window.get_window_type()) {
     case Meta.WindowType.DROPDOWN_MENU:
@@ -278,16 +279,18 @@ LayoutManager.prototype = {
         this.hotCornerManager = null;
         this.edgeRight = null;
         this.edgeLeft = null;
-        this.hideIdleId = 0;
         this._chrome = new Chrome(this);
 
         this.enabledEdgeFlip = global.settings.get_boolean("enable-edge-flip");
         this.edgeFlipDelay = global.settings.get_int("edge-flip-delay");
 
-        this.keyboardBox = new St.BoxLayout({ name: 'keyboardBox',
-                                              reactive: true,
-                                              track_hover: true });
+        this.keyboardBox = new St.Widget({ name: 'keyboardBox',
+                                           layout_manager: new Clutter.BinLayout(),
+                                           important: true,
+                                           reactive: true,
+                                           track_hover: true });
         this.keyboardBox.hide();
+        this._keyboardIndex = -1;
 
         this.addChrome(this.keyboardBox, { visibleInFullscreen: true, affectsStruts: false });
 
@@ -356,11 +359,14 @@ LayoutManager.prototype = {
         if (this.hotCornerManager)
             this.hotCornerManager.update();
         this._chrome._queueUpdateRegions();
+
+        this.keyboardIndex = this.primaryIndex;
     },
 
     _monitorsChanged: function() {
         this._updateMonitors();
         this._updateBoxes();
+        this._updateKeyboardBox()
         this.emit('monitors-changed');
     },
 
@@ -419,53 +425,105 @@ LayoutManager.prototype = {
         Main.setRunState(Main.RunState.RUNNING);
     },
 
-    showKeyboard: function () {
-        if (this.hideIdleId > 0) {
-            Mainloop.source_remove(this.hideIdleId);
-            this.hideIdleId = 0;
+    _updateKeyboardBox: function() {
+        if (Main.panelManager == null || Main.virtualKeyboardManager == null) {
+            return;
         }
 
-        if (!this.keyboardBox.visible) {
-            this.keyboardBox.show();
-        }
+        let size = Main.virtualKeyboardManager.getKeyboardSize();
+        let top = Main.virtualKeyboardManager.getKeyboardPosition() == "top";
+        let panels = Main.panelManager.getPanelsInMonitor(this.keyboardIndex);
 
-        // this.keyboardBox.raise_top();
-        Main.panelManager.lowerActorBelowPanels(this.keyboardBox);
+        let kb_height = this.keyboardMonitor.height / size;
 
-        // Poke Chrome to update the input shape; it doesn't notice
-        // anchor point changes
-        this._chrome.modifyActorParams(this.keyboardBox, { affectsStruts: true });
-        this._chrome.updateRegions();
-
-        this._keyboardHeightNotifyId = this.keyboardBox.connect('notify::height', Lang.bind(this, function () {
-            if (this.keyboardBox.y != 0) {
-                this.keyboardBox.y = this.focusMonitor.y + this.focusMonitor.height - this.keyboardBox.height;
+        let kb_x = this.keyboardMonitor.x;
+        let kb_y = top ? 0 : this.keyboardMonitor.y + this.keyboardMonitor.height - kb_height;
+        let kb_width = this.keyboardMonitor.width;
+        for (let panel of panels) {
+            if (panel.isHideable()) {
+                continue;
             }
-        }));
 
-    },
-
-    queueHideKeyboard: function() {
-        if (this.hideIdleId != 0) {
-            Mainloop.source_remove(this.hideIdleId);
-            this.hideIdleId = 0;
+            switch (panel.panelPosition) {
+                case Panel.PanelLoc.top:
+                    if (top) {
+                        kb_height -= panel.actor.height;
+                        kb_y += panel.actor.height;
+                    }
+                    break;
+                case Panel.PanelLoc.bottom:
+                    if (!top) {
+                        kb_height -= panel.actor.height;
+                    }
+                    break;
+                case Panel.PanelLoc.left:
+                    kb_x += panel.actor.width;
+                    kb_width -= panel.actor.width;
+                    break;
+                case Panel.PanelLoc.right:
+                    kb_width -= panel.actor.width;
+                    break;
+            }
         }
 
-        if (this._keyboardHeightNotifyId) {
-            this.keyboardBox.disconnect(this._keyboardHeightNotifyId);
-            this._keyboardHeightNotifyId = 0;
-        }
-
-        this.hideIdleId = Mainloop.idle_add(Lang.bind(this, this.hideKeyboard));
+        this.keyboardBox.set_position(kb_x, kb_y);
+        this.keyboardBox.set_size(kb_width, kb_height);
     },
 
-    hideKeyboard: function (immediate) {
-        this.keyboardBox.hide();
+    get keyboardMonitor() {
+        return this.monitors[this.keyboardIndex];
+    },
+
+    set keyboardIndex(v) {
+        this._keyboardIndex = v;
+        this._updateKeyboardBox();
+    },
+
+    get keyboardIndex() {
+        return this._keyboardIndex;
+    },
+
+    showKeyboard: function() {
+        this.keyboardBox.opacity = 0;
+        this.keyboardBox.show();
+        this.keyboardBox.remove_all_transitions();
+
+        this.keyboardBox.ease({
+            opacity: 255,
+            duration: KEYBOARD_FADE_TIME,
+            mode: Clutter.AnimationMode.EASE_OUT_QUAD,
+            onComplete: () => {
+                this._showKeyboardComplete();
+            }
+        });
+    },
+
+    _showKeyboardComplete: function() {
+        this._chrome.modifyActorParams(this.keyboardBox, { affectsStruts: true });
+        this._chrome._queueUpdateRegions();
+        this.emit('keyboard-visible-changed', true);
+    },
+
+    hideKeyboard: function(immediate) {
+        this.keyboardBox.remove_all_transitions();
         this._chrome.modifyActorParams(this.keyboardBox, { affectsStruts: false });
-        this._chrome.updateRegions();
+        this._chrome._queueUpdateRegions();
 
-        this.hideIdleId = 0;
-        return false;
+        this.keyboardBox.ease({
+            opacity: 0,
+            duration: immediate ? 0 : KEYBOARD_FADE_TIME,
+            mode: Clutter.AnimationMode.EASE_IN_QUAD,
+            onComplete: () => {
+                this._hideKeyboardComplete();
+            }
+        });
+
+        this.emit('keyboard-visible-changed', false);
+    },
+
+    _hideKeyboardComplete: function() {
+        this.keyboardBox.hide();
+        this.keyboardBox.opacity = 255;
     },
 
     /**
@@ -979,9 +1037,11 @@ Chrome.prototype = {
                 let monitor = this.findMonitorForActor(actorData.actor);
                 let side;
                 if (x1 <= monitor.x && x2 >= monitor.x + monitor.width) {
-                    if (y1 <= monitor.y)
+                    // Allow the keyboard box to become chrome even though it's only adjacent to a panel
+                    // rather than a monitor edge.
+                    if (y1 <= monitor.y + (actorData.actor.name === "keyboardBox" ? 100 : 0))
                         side = Meta.Side.TOP;
-                    else if (y2 >= monitor.y + monitor.height)
+                    else if (y2 >= monitor.y + monitor.height - (actorData.actor.name === "keyboardBox" ? 100 : 0))
                         side = Meta.Side.BOTTOM;
                     else
                         continue;
